from numpy.testing import assert_equal


def function_test(dask_trj, md_trj, funcname, **kwargs):
    if '.' in funcname:
        funcmod, funcn = funcname.rsplit('.', 1)
        dmdfuncmod = 'dask_traj.{}'.format(funcmod)
        mdfuncmod = 'mdtraj.{}'.format(funcmod)
    else:
        funcn = funcname
        dmdfuncmod = 'dask_traj'
        mdfuncmod = 'mdtraj'

    dmdfunc = getattr(__import__(dmdfuncmod), funcn)
    mdfunc = getattr(__import__(mdfuncmod), funcn)
    dmdr = dmdfunc(dask_trj, **kwargs).compute()
    mdr = mdfunc(md_trj, **kwargs)
    assert_equal(dmdr, mdr)



try:
    from . import version
except ImportError:  # pragma: no cover
    from . import _version as version

__version__ = version.version

from .utils import ensure_type
from .geometry import *
from .core import *



import pytest
import os
from dask_traj.core.dask_traj import (build_result_dict,
                                      get_xyz,
                                      get_time,
                                      get_unitcell,
                                      read_chunk,
                                      load)
import dask
import numpy as np
import mdtraj as md

rel_dir = os.path.dirname(__file__)
xtc = os.path.join(rel_dir, 'test.xtc')
pdb = os.path.join(rel_dir, 'test.pdb')
txt = os.path.join(rel_dir, 'test.xml')
xyz = os.path.join(rel_dir, 'test.xyz')


def f(x):
    return np.asarray(x)

class TestDaskTraj(object):
    def setup(self):
        pass

    def teardown(self):
        pass

    def test_not_implemented_filetype(self):
        with pytest.raises(NotImplementedError):
            build_result_dict(txt, extension='.xml',
                              length=1, chunk_size=1,
                              distance_unit='nm')

    def test_no_xyz_in_result_dict(self):
        assert get_xyz(result_dict={}, length=1, distance_unit='nm') is None

    def test_no_time_in_result_dict(self):
        time = get_time(result_dict={},
                        length=10,
                        chunk_size=2)
        assert list(time) == [0,1,2,3,4,5,6,7,8,9]

    def test_no_unitcell_in_result_dict(self):
        assert get_unitcell(result_dict={}, length=1) == (None, None, None)

    def test_no_uv_and_ua_in_result_dict(self):
        rd = {'unitcell_lengths':[dask.delayed(f)([1,2,3])]}
        ul, ua, uv = get_unitcell(result_dict=rd, length=3)
        assert uv is None
        assert (ul.compute() == np.asarray([1,2,3])).all()
        assert (ua.compute() == np.asarray([1,1,1])).all()

    def test_no_ul_in_result_dict(self):
        rd = {'unitcell_angles':[dask.delayed(f)([1,2,3])],
              'unitcell_vectors':[dask.delayed(f)([3,4,5])]}
        ul, ua, uv = get_unitcell(result_dict=rd, length=3)
        assert (uv.compute() == np.asarray([3,4,5])).all()
        assert (ul.compute() == np.asarray([3,4,5])).all()
        assert (ua.compute() == np.asarray([1,2,3])).all()

    def test_chunk_with_one_return(self):
        assert type(read_chunk(xyz, 1, 0)) is tuple

    def test_trajectory_to_mdtraj(self):
        dask_traj = load(xtc, top=pdb)
        md_traj = md.load(xtc, top=pdb)
        dask_traj.to_mdtraj()
        assert dask_traj == md_traj

    def test_trajectory_to_mdtraj_with_unitcell_lengths(self):
        dask_traj = load(xtc, top=pdb)
        #Trigger the calculation of unitcell lengths
        _ = dask_traj.unitcell_lengths
        md_traj = md.load(xtc, top=pdb)
        dask_traj.to_mdtraj()
        assert dask_traj == md_traj

    def test_set_xyz_without_top(self):
        dask_traj = load(xtc, top=pdb)
        dask_traj.top = None
        assert dask_traj.top is None
        dask_traj.xyz = np.array([[[1.,2.,3.]]], dtype=np.float32)
        assert (dask_traj._xyz == np.array([[[1.,2.,3.]]])).all()

    def test_calc_uv_without_ul_or_ua(self):
        dask_traj = load(xtc, top=pdb)
        dask_traj.unitcell_vectors = None
        assert dask_traj._calc_unitcell_vectors() is None

    def test_wrong_uv_length(self):
        dask_traj = load(xtc, top=pdb)
        with pytest.raises(TypeError) as err:
            dask_traj._calc_length_and_angles([1,2,3])

    def test_join(self):
         dt = load(xtc, top=pdb)
         dt2 = load(xtc, top=pdb)
         dt3 = dt.join(dt2)
         assert len(dt3) == len(dt)+len(dt2)

    def test_join_failures(self):
         dt = load(xtc, top=pdb)
         dt2 = 'not a trajectory'

         # Not a trajectory
         with pytest.raises(TypeError) as err:
             _ = dt.join(dt2)
         assert 'must be a list of Trajectory' in str(err.value)

         with pytest.raises(TypeError) as err:
             _ = dt.join([dt2])
         assert 'only join Trajectory instances' in str(err.value)

         dt2 = load(xtc, top=pdb)
         dt2.top = None # Remove topology
         with pytest.raises(ValueError) as err:
             _ = dt.join([dt2])
         assert 'The topologies of the Trajectories are not' in str(err.value)

         dt2 = load(xtc, top=pdb)
         _ = dt2.unitcell_lengths
         dt2.atom_slice([1,2,3], inplace=True)
         with pytest.raises(ValueError) as err:
             _ = dt.join([dt2])
         assert 'Number of atoms' in str(err.value)


         dt2 = load(xtc, top=pdb)
         dt2.unitcell_vectors = None # Remove unitcell
         with pytest.raises(ValueError) as err:
             _ = dt.join([dt2])
         assert 'unitcell' in str(err.value)

    def test_join_discard_overlap(self):
         dt = load(xtc, top=pdb)
         dt2 = load(xtc, top=pdb)
         dt3 = dt.join(dt2[::-1], discard_overlapping_frames=True)
         assert len(dt3) == len(dt)+len(dt2)-1

    def test_hash(self):
         dt = load(xtc, top=pdb)
         dt2 = load(xtc, top=pdb)
         assert dt == dt2 # This will compare hashes




import dask
import dask.array as da
from dask_traj.utils import ensure_type, wrap_da
from dask_traj.geometry import distance
from mdtraj.geometry import _geometry
import numpy as np


@dask.delayed
def _compute_angles_chunk(xyz, triplets, box=None, periodic=True,
                          opt=True, orthogonal=False):
    """Compute the angles for a single chunk

    Parameters
    ----------
    xyz : ndarray of shape (any, any, 3)
        The xyz coordinates of the chunk
    triplets : array of shape (any, 3)
        The indices for which to compute an angle
    box : ndarray of shape (any, 3, 3)
        The box vectors of the chunk
    periodic : bool
        Wether to use the periodc boundary during the calculation.
    opt : bool, default=True
        Use an optimized native library to calculate distances. MDTraj's
        optimized SSE angle calculation implementation is 10-20x faster than
        the (itself optimized) numpy implementation.
    orthogonal : bool or da.bool
        Wether all angles are close to 90 degrees
    """
    # Cast dask.bool to a true bool
    orthogonal = bool(orthogonal)
    xyz = ensure_type(xyz, dtype=np.float32, ndim=3, name='xyz',
                      shape=(None, None, 3), warn_on_cast=False,
                      cast_da_to_np=True)

    out = np.empty((xyz.shape[0], triplets.shape[0]), dtype=np.float32)
    if opt:
        if periodic and box is not None:
            _geometry._angle_mic(xyz, triplets, box.transpose(0, 2, 1).copy(),
                                 out, orthogonal)
        else:
            _geometry._angle(xyz, triplets, out)
    else:
        out = _angle(xyz, triplets, periodic, out).compute()
    return out


def compute_angles(traj, angle_indices, periodic=True, opt=True, **kwargs):
    """ Daskified version of mdtraj.compute_angles().

    This mimics py:method:`mdtraj.compute_angles()` but returns the answer
    as a py:class:`dask.array` object

    Parameters
    ----------
    traj : :py:class:`dask_traj.Trajectory`
        The trajectory to compute the angles for.
    angle_indices : array of shape(any, 3)
        The indices for which to compute an angle.
    periodic : bool
        Wether to use the periodc boundary during the calculation.
    opt : bool, default=True
        Use an optimized native library to calculate distances. MDTraj's
        optimized SSE angle calculation implementation is 10-20x faster than
        the (itself optimized) numpy implementation.

    Returns
    -------
    angles : dask.array, shape(n_frames, angle_indices)
        Dask array with the delayed calculated angle for each item in
        angle_indices for each frame.
    """

    xyz = traj.xyz
    length = len(xyz)
    atoms = len(angle_indices)
    triplets = ensure_type(angle_indices, dtype=np.int32, ndim=2,
                           name='angle_indices', shape=(None, 3),
                           warn_on_cast=False)
    if not np.all(np.logical_and(triplets < traj.n_atoms, triplets >= 0)):
        raise ValueError('angle_indices must be between 0 and %d' %
                         traj.n_atoms)

    if len(triplets) == 0:
        return da.zeros((len(xyz), 0), dtype=np.float32)

    if periodic and traj._have_unitcell:
        box = ensure_type(traj.unitcell_vectors, dtype=np.float32, ndim=3,
                          name='unitcell_vectors', shape=(len(xyz), 3, 3),
                          warn_on_cast=False)
    else:
        box = None
        orthogonal = False

    lazy_results = []
    current_frame = 0
    for frames in xyz.chunks[0]:
        next_frame = current_frame+frames
        if box is not None:
            current_box = box[current_frame:next_frame]
            orthogonal = da.allclose(
                traj.unitcell_angles[current_frame:next_frame], 90)
        else:
            current_box = None
        chunk_size = (frames, atoms)
        lazy_results.append(wrap_da(_compute_angles_chunk, chunk_size,
                                    xyz=xyz[current_frame:next_frame],
                                    triplets=triplets,
                                    box=current_box,
                                    orthogonal=orthogonal,
                                    opt=opt,
                                    **kwargs))
        current_frame = next_frame
    max_result = da.concatenate(lazy_results)
    result = max_result[:length]
    return result


def _angle(xyz, triplets, periodic, out):
    """Delayed version of the _angle function of mdtraj

    Parameters
    ----------
    xyz : ndarray of shape (any, any, 3)
        The xyz coordinates of the chunk
    triplets : array of shape (any, 3)
        The indices for which to compute an angle
    box : ndarray of shape (any, 3, 3)
        The box vectors of the chunk
    periodic : bool
        Wether to use the periodc boundary during the calculation.
    out : ndarray of zeros of shape (len(xyz), len(triplets))
        The output array
    """

    ix01 = triplets[:, [1, 0]]
    ix21 = triplets[:, [1, 2]]

    u_prime = distance._compute_displacements_chunk(xyz, ix01,
                                                    periodic=periodic,
                                                    opt=False)
    v_prime = distance._compute_displacements_chunk(xyz, ix21,
                                                    periodic=periodic,
                                                    opt=False)
    u_norm = ((u_prime**2).sum(-1))**(1/2)
    v_norm = ((v_prime**2).sum(-1))**(1/2)

    # adding a new axis makes sure that broasting rules kick in on the third
    # dimension
    u = u_prime / (u_norm[..., np.newaxis])
    v = v_prime / (v_norm[..., np.newaxis])

    return dask.delayed(np.arccos)((u * v).sum(-1), out=out)



import dask
import dask.array as da
import numpy as np
import mdtraj
import os
from mdtraj.utils import in_units_of
from mdtraj.core.trajectory import (open, _get_extension, _parse_topology,
                                    _TOPOLOGY_EXTS, _hash_numpy_array)
from dask_traj.utils import (ensure_type, box_vectors_to_lengths_and_angles,
                             lengths_and_angles_to_box_vectors)

from copy import deepcopy
# dictionary to tell what is actually returned by read per extension type
# can consist of [xyz, time, unitcell_lengths, unitcell_angles,
# unitcell_vectors]
file_returns = {'.arc': ['xyz', 'unitcell_lengths', 'unitcell_angles'],
                '.dcd': ['xyz', 'unitcell_lengths', 'unitcell_angles'],
                '.binpos': ['xyz'],
                '.xtc': ['xyz', 'time', 'step', 'unitcell_vectors'],
                '.trr': ['xyz', 'time', 'step', 'unitcell_vectors', '_'],
                '.hdf5': ['data'],  # Need special case
                '.h5': ['data'],  # Same as hdf5
                '.ncdf': ['xyz', 'time', 'unitcell_lengths',
                          'unitcell_angles'],
                '.netcdf': ['xyz', 'time', 'unitcell_lengths',
                            'unitcell_angles'],
                '.nc': ['xyz', 'time', 'unitcell_lengths',
                        'unitcell_angles'],
                '.pdb.gz': [],  # Not implemented for now
                '.pdb': [],  # Not implemented for now
                '.lh5': ['xyz'],
                '.crd': ['xyz', 'unitcell_lengths'
                         ],  # Needs to assume angles to be 90
                '.mdcrd': ['xyz', 'unitcell_lengths'
                           ],  # Needs to assume angles to be 90
                '.inpcrd': ['xyz', 'time', 'unitcell_lengths',
                            'unitcell_angles'],
                '.restrt': ['xyz', 'time', 'unitcell_lengths',
                            'unitcell_angles'],
                '.rst7': ['xyz', 'time', 'unitcell_lengths',
                          'unitcell_angles'],
                '.ncrst': ['xyz', 'time', 'unitcell_lengths',
                           'unitcell_angles'],
                '.lammpstrj': ['xyz', 'unitcell_lengths', 'unitcell_angles'],
                '.dtr': ['xyz', 'time', 'unitcell_lengths', 'unitcell_angles'],
                '.stk': ['xyz', 'time', 'unitcell_lengths', 'unitcell_angles'],
                '.gro': ['xyz', 'time', 'unitcell_vectors'],
                '.xyz.gz': ['xyz'],
                '.xyz': ['xyz'],
                '.tng': ['xyz', 'time', 'unitcell_vectors'],
                '.xml': [],  # not implemented for now
                '.mol2': [],  # not implemented for now
                '.hoomdxml': []  # not implemented for now
                }

# TODO make class of trajectory chunk


def load(filename, chunks=10, **kwargs):
    """
    A loader that will mimic :py:func:`mdtraj.Trajectory.load()`, but
    construct a :py:class:`dasktraj.Trajectory` with a :py:class:`dask.array`
    as xyz, time, and unitcell properties.

    Parameters
    ----------
    filename : string
        Filename of the file to load.
    chunks : int
        Number of frames per chunk.

    Returns
    -------
    trajectory
        A :py:class:`dasktraj.Trajectory`
    """

    top = kwargs.pop('top', None)
    extension = _get_extension(filename)
    if extension not in _TOPOLOGY_EXTS:
        topology = _parse_topology(top)

    filename = os.path.abspath(filename)
    length = len(open(filename))
    n_chunks = int(length/chunks)
    frames_left = length % chunks
    if frames_left != 0:
        n_chunks += 1
    # TODO this needs to be closed at some point
    data = load_chunks(filename, extension, chunks, range(n_chunks),
                       **kwargs)

    # TODO: use this to construct unitcells
    # Pop out irrelevant info
    uv = data.pop('unitcell_vectors')
    traj = Trajectory(topology=topology, delayed_objects=data, **data)
    if uv is not None:
        traj.unitcell_vectors = uv
    return traj


def load_chunks(filename, extension, chunk_size, chunks, **kwargs):
    """
    Uses :py:func:`dask.delayed()` to lazy load the chunks of a trajectory
    and builds a result dict.

    Parameters
    ----------
    filename : string
        Filename of the file to load.
    extension: string
        File extension of the file to load.
    chunk_size: int
        Number of frames per chunk.
    chunks: list of int
        Which chunks to load


    Returns
    -------
    dict
        A dictionary with the :py:class:`dask.array` trajectory properties
    """

    # TODO: Add other kwargs like stride, atom_indices, time

    with open(filename) as f:
        length = len(f)
        distance_unit = f.distance_unit
    results = []
    for chunk in chunks:
        start = chunk*chunk_size
        frames = min(length-start, chunk_size)
        results.append(dask.delayed(read_chunk, pure=True)(filename,
                                                           frames,
                                                           start))

    result_dict = build_result_dict(results, extension, length, chunk_size,
                                    distance_unit)
    return result_dict


def build_result_dict(results, extension, length, chunk_size, distance_unit):
    """
    Builds a result dict with py:class:`dask.array` for the properties of the
    trajectory

    Parameters
    ----------
    results : list of :py:class:`dask.Delayed` objects
        delayed objects to extract the trajectory properties from.
    extension: string
        File extension of the file to load.
    lenght: int
        Number of frames of the trajectory.
    chunk_size: int
        Number of frames per chunk.
    distance_unit: string
        Indication of the distance unit of the file for the xyz.

    Returns
    -------
    dict
        A dictionary with the :py:class:`dask.array` trajectory properties
    """

    read_returns = file_returns[extension]
    if len(read_returns) == 0:
        raise NotImplementedError(
            "{} files are not yet supported".format(extension))

    # Persis the sample for quick building
    _ = results[0].persist()
    result_dict = {key: [result[i] for result in results]
                   for i, key in enumerate(read_returns)}
    xyz = get_xyz(result_dict, length, distance_unit)
    time = get_time(result_dict, length, chunk_size)
    unit_cell = get_unitcell(result_dict, length)

    not_xyz_dict = {'time': time,
                    'unitcell_lengths': unit_cell[0],
                    'unitcell_angles': unit_cell[1],
                    'unitcell_vectors': unit_cell[2]}
    return_dict = not_xyz_dict
    return_dict['xyz'] = xyz
    return return_dict


def make_da(delayed_list, length):
    """
    Makes an  py:class:`dask.array` from a list of delayed objects.

    Parameters
    ----------
    delayed_list: list of :py:class:`dask.delayed` objects
        list of delayed objects to convert into an dask array
    lenght : int
        total length of the final dask array.

    Returns
    -------
    :py:class:`dask.array`
        dask array from the delayed objects.
    """
    sample = delayed_list[0].compute()
    arrays = [da.from_delayed(item, dtype=sample.dtype, shape=sample.shape)
              for item in delayed_list]
    result = da.concatenate(arrays, axis=0)[:length]
    return result


def get_xyz(result_dict, length, distance_unit):
    """
    Makes an  py:class:`dask.array` for xyz if it can be loaded from the
    fileformat, otherwise returns None.

    Parameters
    ----------
    result_dict: dict of :py:class:`dask.delayed` objects
        dict of delayed objects where we make the xyz from into a dask   array.
    lenght : int
        total length of the final dask array.
    distance_unit: string
        distance unit of the filetype to be loaded.

    Returns
    -------
    :py:class:`dask.array` or None
        dask array from the delayed objects for xyz if it can be loaded,
        None otherwise.
    """

    xyz_list = result_dict.pop('xyz', None)
    if xyz_list is None:
        return None
    else:
        for xyz in xyz_list:
            in_units_of(xyz, distance_unit, Trajectory._distance_unit,
                        inplace=True)
    result = make_da(xyz_list, length)
    return result


def get_time(result_dict, length, chunk_size):
    """
    Makes an  py:class:`dask.array` for time if it can be loaded from the
    fileformat, otherwise returns None.

    Parameters
    ----------
    result_dict: dict of :py:class:`dask.delayed` objects
        dict of delayed objects where we make the time from into an dask array.
    lenght : int
        total length of the final dask array.
    chunk_size: int
        length of a single chunk.

    Returns
    -------
    :py:class:`dask.array` or None
        dask array from the delayed objects for time if it can be loaded,
        None otherwise.
    """
    time_list = result_dict.pop('time', None)
    if time_list is None:
        # TODO incorporate stride
        result = da.arange(length, chunks=(chunk_size,))
    else:
        result = make_da(time_list, length)
    return result


def get_unitcell(result_dict, length):
    """
    Makes an  py:class:`dask.array` for the unitcell information if it can be
    loaded from the fileformat, otherwise returns None.

    Parameters
    ----------
    result_dict: dict of :py:class:`dask.delayed` objects
        dict of delayed objects where we make the time from  grab  into an dask
        array.
    lenght : int
        total length of the final dask array.

    Returns
    -------
    (unitcell_lengths, unitcell_angles, unitcell_vectors)
        dask array from the delayed objects for each term if it can be loaded,
        None otherwise.
    """

    # TODO add ensure type on these lengths
    unitcell_lengths = result_dict.pop('unitcell_lengths', None)
    unitcell_angles = result_dict.pop('unitcell_angles', None)
    unitcell_vectors = result_dict.pop('unitcell_vectors', None)
    if (unitcell_lengths is None and unitcell_angles is None and
            unitcell_vectors is None):
        ul = None
        ua = None
        uv = None
        return ul, ua, uv

    # Now if there is info to be loaded
    if unitcell_vectors is not None:
        uv = make_da(unitcell_vectors, length)
    else:
        uv = None

    if unitcell_lengths is not None and unitcell_angles is None:
        ul = make_da(unitcell_lengths, length)
        ua = da.ones_like(ul)
    elif unitcell_angles is not None:
        ul = make_da(unitcell_vectors, length)
        ua = make_da(unitcell_angles, length)
    else:
        ul = None
        ua = None

    return ul, ua, uv


def read_chunk(filename, chunk_size, start):
    """
    Read a chunk of the trajectory file.

    Parameters
    ----------
    filename: string
        filename of the file to load
    chunk_size: int
        size of the chunk to load
    start: int
        frame to start the loading from

    Returns
    -------
    A tuple with the data for the read chunk (depending on what
    filetype.read returns)
    """

    with open(filename) as f:
        # Get current possition
        pos = f.tell()
        # position we want
        seek_pos = start
        rel_pos = seek_pos-pos

        f.seek(rel_pos, 1)
        result = f.read(chunk_size)

    if isinstance(result, tuple):
        return result
    else:
        return result,


class Trajectory(mdtraj.Trajectory):
    """mdtraj.Trajectory with dask.array properties

    Attributes
    ----------
    n_frames : int
    n_atoms : int
    n_residues : int
    time : np.ndarray, shape=(n_frames,)
    timestep : float
    topology : md.Topology
    top : md.Topology
    xyz : ndarray, shape=(n_frames, n_atoms, 3)
    unitcell_vectors : {ndarray, shape=(n_frames, 3, 3), None}
    unitcell_lengths : {ndarray, shape=(n_frames, 3), None}
    unitcell_angles : {ndarray, shape=(n_frames, 3), None}
    """

    # TODO add other kwargs from MDtraj.trajectory
    def __init__(self, xyz,  topology, time=None, delayed_objects=None,
                 **kwargs):
        dask.persist(**kwargs)
        self._unitcell_vectors = None
        super(Trajectory, self).__init__(xyz=xyz, topology=topology, time=time,
                                         **kwargs)

    def to_mdtraj(self):
        """Converts self to a mdtraj.Trajectory"""

        self.xyz = np.asarray(self.xyz)
        self.time = np.asarray(self.time)
        if self._have_unitcell and self._unitcell_lengths is None:
            self._calc_length_and_angles(self.unitcell_vectors)
            self.unitcell_lengths = np.asarray(self.unitcell_lengths)
            self.unitcell_angles = np.asarray(self.unitcell_angles)
            # We don't set the vectors here as they are calculated in
            # mdtraj.Trajectory anyway
        elif self._have_unitcell:
            self.unitcell_lengths = np.asarray(self.unitcell_lengths)
            self.unitcell_angles = np.asarray(self.unitcell_angles)
        self.__class__ = mdtraj.Trajectory

    @property
    def xyz(self):
        """Cartesian coordinates of each atom in each simulation frame

        Returns
        -------
        xyz : ndarray, shape=(n_frames, n_atoms, 3)
            A three dimensional dask array, with the cartesian coordinates
            of each atoms in each frame.
        """
        return self._xyz

    @xyz.setter
    def xyz(self, value):
        "Set the cartesian coordinates of each atom in each simulation frame"
        if self.top is not None:
            # if we have a topology and its not None
            shape = (None, self.topology._numAtoms, 3)
        else:
            shape = (None, None, 3)

        value = ensure_type(value, np.float32, 3, 'xyz', shape=shape,
                            warn_on_cast=True,
                            add_newaxis_on_deficient_ndim=True)
        self._xyz = value
        self._rmsd_traces = None

    @property
    def _have_unitcell(self):
        return ((self._unitcell_lengths is not None and
                 self._unitcell_angles is not None) or
                self._unitcell_vectors is not None)

    @property
    def unitcell_angles(self):
        """Angles that define the shape of the unit cell in each frame.

        Returns
        -------
        lengths : nd.array, shape=(n_frames, 3)
            The angles between the three unitcell vectors in each frame,
            ``alpha``, ``beta``, and ``gamma``. ``alpha' gives the angle
            between vectors ``b`` and ``c``, ``beta`` gives the angle between
            vectors ``c`` and ``a``, and ``gamma`` gives the angle between
            vectors ``a`` and ``b``. The angles are in degrees.
        """
        if self._unitcell_angles is None:
            self._calc_length_and_angles(self._unitcell_vectors)
        return self._unitcell_angles

    @unitcell_angles.setter
    def unitcell_angles(self, value):
        """Set the lengths that define the shape of the unit cell in each frame

        Parameters
        ----------
        value : ndarray, shape=(n_frames, 3)
            The angles ``alpha``, ``beta`` and ``gamma`` that define the
            shape of the unit cell in each frame. The angles should be in
            degrees.
        """
        self._unitcell_angles = ensure_type(value, np.float32, 2,
                                            'unitcell_angles',
                                            can_be_none=True,
                                            shape=(len(self), 3),
                                            warn_on_cast=False,
                                            add_newaxis_on_deficient_ndim=True)

    @property
    def unitcell_lengths(self):
        """Lengths that define the shape of the unit cell in each frame.

        Returns
        -------
        lengths : {ndarray, shape=(n_frames, 3), None}
            Lengths of the unit cell in each frame, in nanometers, or None
            if the Trajectory contains no unitcell information.
        """
        if self._unitcell_lengths is None:
            self._calc_length_and_angles(self._unitcell_vectors)
        return self._unitcell_lengths

    @unitcell_lengths.setter
    def unitcell_lengths(self, value):
        """Set the lengths that define the shape of the unit cell in each frame

        Parameters
        ----------
        value : ndarray, shape=(n_frames, 3)
            The distances ``a``, ``b``, and ``c`` that define the shape of the
            unit cell in each frame, or None
        """
        self._unitcell_lengths = ensure_type(value, np.float32, 2,
                                             'unitcell_lengths',
                                             can_be_none=True,
                                             shape=(len(self), 3),
                                             warn_on_cast=False,
                                             add_newaxis_on_deficient_ndim=True
                                             )

    @property
    def unitcell_vectors(self):
        """vectors that define the shape of the unit cell in each frame.

        Returns
        -------
        vectors : {ndarray, shape=(n_frames, 3), None}
            Vecotors of the unit cell in each frame, in nanometers, or None
            if the Trajectory contains no unitcell information.
        """

        if self._unitcell_vectors is None:
            self._unitcell_vectors = self._calc_unitcell_vectors()
        return self._unitcell_vectors

    def _calc_unitcell_vectors(self):
        """The vectors that define the shape of the unit cell in each frame

        Returns
        -------
        vectors : da.ndarray, shape(n_frames, 3, 3)
            Vectors defining the shape of the unit cell in each frame.
            The semantics of this array are that the shape of the unit cell
            in frame ``i`` are given by the three vectors, ``value[i, 0, :]``,
            ``value[i, 1, :]``, and ``value[i, 2, :]``.
        """
        if self.unitcell_lengths is None or self.unitcell_angles is None:
            return None

        v1, v2, v3 = lengths_and_angles_to_box_vectors(
            self._unitcell_lengths[:, 0],  # a
            self._unitcell_lengths[:, 1],  # b
            self._unitcell_lengths[:, 2],  # c
            self._unitcell_angles[:, 0],   # alpha
            self._unitcell_angles[:, 1],   # beta
            self._unitcell_angles[:, 2],   # gamma
        )
        return da.swapaxes(da.dstack((v1, v2, v3)), 1, 2)

    @unitcell_vectors.setter
    def unitcell_vectors(self, vectors):
        self._unitcell_vectors = vectors

    def _calc_length_and_angles(self, vectors):
        """Set the three vectors that define the shape of the unit cell

        Parameters
        ----------
        vectors : tuple of three arrays, each of shape=(n_frames, 3)
            The semantics of this array are that the shape of the unit cell
            in frame ``i`` are given by the three vectors, ``value[i, 0, :]``,
            ``value[i, 1, :]``, and ``value[i, 2, :]``.
        """
        if vectors is None:  # or da.all(abs(vectors) < 1e-15):
            self._unitcell_lengths = None
            self._unitcell_angles = None
            return

        if not len(vectors) == len(self):
            raise TypeError('unitcell_vectors must be the same length as '
                            'the trajectory. you provided %s' % str(vectors))

        v1 = vectors[:, 0, :]
        v2 = vectors[:, 1, :]
        v3 = vectors[:, 2, :]
        a, b, c, alpha, beta, gamma = box_vectors_to_lengths_and_angles(v1,
                                                                        v2,
                                                                        v3)

        self._unitcell_lengths = da.vstack((a, b, c)).T
        self._unitcell_angles = da.vstack((alpha, beta, gamma)).T

    def join(self, other, check_topology=True,
             discard_overlapping_frames=False):
        """This is a daskified version of md.Trajectory.join()

        Parameters
        ----------
        other : Trajectory or list of Trajectory
            One or more trajectories to join with this one. These trajectories
            are *appended* to the end of this trajectory.
        check_topology : bool
            Ensure that the topology of `self` and `other` are identical before
            joining them. If false, the resulting trajectory will have the
            topology of `self`.
        discard_overlapping_frames : bool, optional
            If True, compare coordinates at trajectory edges to discard
            overlapping frames.  Default: False.
        """

        if isinstance(other, Trajectory):
            other = [other]
        if isinstance(other, list):
            if not all(isinstance(o, Trajectory) for o in other):
                raise TypeError('You can only join Trajectory instances')
            if not all(self.n_atoms == o.n_atoms for o in other):
                raise ValueError('Number of atoms in self (%d) is not equal '
                                 'to number of atoms in other' %
                                 (self.n_atoms))
            if check_topology and not all(self.topology == o.topology
                                          for o in other):
                raise ValueError('The topologies of the Trajectories are not '
                                 'the same')
            if not all(self._have_unitcell == o._have_unitcell for o in other):
                raise ValueError('Mixing trajectories with and without '
                                 'unitcell')
        else:
            raise TypeError(
                '`other` must be a list of Trajectory. You supplied %s' %
                type(other))

        trajectories = [self] + other
        if discard_overlapping_frames:
            for i in range(len(trajectories)-1):
                # last frame of trajectory i
                x0 = trajectories[i].xyz[-1]
                # first frame of trajectory i+1
                x1 = trajectories[i + 1].xyz[0]

                # check that all atoms are within 2e-3 nm
                # (this is kind of arbitrary)
                if np.all(np.abs(x1 - x0) < 2e-3):
                    trajectories[i] = trajectories[i][:-1]

        # Only difference between original code and current code
        xyz = da.concatenate([t.xyz for t in trajectories])

        time = np.concatenate([t.time for t in trajectories])
        angles = lengths = None
        if self._have_unitcell:
            angles = np.concatenate([t.unitcell_angles for t in trajectories])
            lengths = np.concatenate([t.unitcell_lengths for
                                      t in trajectories])

        # use this syntax so that if you subclass Trajectory,
        # the subclass's join() will return an instance of the subclass
        return self.__class__(xyz, deepcopy(self._topology), time=time,
                              unitcell_lengths=lengths,
                              unitcell_angles=angles)

    def __hash__(self):
        '''Updated hash function to use the name of the dask arrays'''
        hash_value = hash(self.top)
        # combine with hashes of arrays
        hash_value ^= hash(self._xyz.name)
        hash_value ^= hash(self.time.name)
        hash_value ^= _hash_numpy_array(self.unitcell_lengths.compute())
        hash_value ^= _hash_numpy_array(self.unitcell_angles.compute())
        return hash_value



import os
from .helpers import function_test
import mdtraj as md
import dask_traj as dmd
import itertools as itt
import pytest
import numpy as np

rel_dir = os.path.dirname(__file__)
xtc = os.path.join(rel_dir, 'test.xtc')
pdb = os.path.join(rel_dir, 'test.pdb')


class TestGeometry(object):
    def setup(self):
        self.md_traj = md.load(xtc, top=pdb)
        self.dask_traj = dmd.load(xtc, top=pdb)
        self.indices = range(self.md_traj.n_atoms)

    def teardown(self):
        pass

    def slice_trajs(self, opt):
        if opt:
            return self.dask_traj, self.md_traj
        else:
            return self.dask_traj[:5], self.md_traj[:5]

    @pytest.mark.parametrize('periodic', [True, False])
    @pytest.mark.parametrize('opt', [True, False])
    def test_compute_distances(self, opt, periodic):
        dask_trj, md_trj = self.slice_trajs(opt)

        c = list(itt.combinations(self.indices[:10], 2))
        function_test(dask_trj=dask_trj,
                      md_trj=md_trj,
                      funcname='geometry.compute_distances',
                      atom_pairs=c,
                      opt=opt,
                      periodic=periodic)

    def test_compute_distances_wrong_pairs(self):
        dask_trj = self.dask_traj
        c = [[12345, -1]]
        with pytest.raises(ValueError) as err:
            dmd.compute_distances(traj=dask_trj,
                                  atom_pairs=c)
        assert 'atom_pairs must be between 0' in str(err.value)

    def test_compute_distances_empty_pairs(self):
        c = np.array([[12345, -1]])
        c = c[0:0]  # Make an empty slice
        function_test(dask_trj=self.dask_traj,
                      md_trj=self.md_traj,
                      funcname='compute_distances',
                      atom_pairs=c)

    @pytest.mark.parametrize('periodic', [True, False])
    @pytest.mark.parametrize('opt', [True, False])
    def test_compute_displacements(self, opt, periodic):
        c = list(itt.combinations(self.indices[:10], 2))
        function_test(dask_trj=self.dask_traj,
                      md_trj=self.md_traj,
                      funcname='compute_displacements',
                      atom_pairs=c,
                      opt=opt,
                      periodic=periodic)

    def test_compute_displacements_wrong_pairs(self):
        dask_trj = self.dask_traj
        c = [[12345, -1]]
        with pytest.raises(ValueError) as err:
            dmd.compute_displacements(traj=dask_trj,
                                      atom_pairs=c)
        assert 'atom_pairs must be between 0' in str(err.value)

    # Bug in mdtraj should pass once https://github.com/mdtraj/mdtraj/pull/1515
    # is merged in. Then this should be removed.
    @pytest.mark.xfail(strict=True)
    def test_compute_displacements_empty_pairs(self):
        c = np.array([[1234, -9999]])
        c = c[0:0]
        function_test(dask_trj=self.dask_traj,
                      md_trj=self.md_traj,
                      funcname='compute_displacements',
                      atom_pairs=c)

    def test_compute_center_of_mass(self):
        function_test(dask_trj=self.dask_traj,
                      md_trj=self.md_traj,
                      funcname='compute_center_of_mass')

    def test_compute_center_of_geometry(self):
        function_test(dask_trj=self.dask_traj,
                      md_trj=self.md_traj,
                      funcname='compute_center_of_geometry')

    @pytest.mark.parametrize('periodic', [True, False])
    def test_find_closest_contact(self, periodic):
        a = self.indices[:10]
        b = self.indices[10:20]
        function_test(dask_trj=self.dask_traj,
                      md_trj=self.md_traj,
                      funcname='find_closest_contact',
                      group1=a,
                      group2=b,
                      periodic=periodic)

    def test_find_closest_contacts_empty_slice(self):
        a = self.indices[10:10]
        b = self.indices[10:20]
        function_test(dask_trj=self.dask_traj,
                      md_trj=self.md_traj,
                      funcname='find_closest_contact',
                      group1=a,
                      group2=b)

    @pytest.mark.parametrize('periodic', [True, False])
    @pytest.mark.parametrize('opt', [True, False])
    def test_compute_angles(self, opt, periodic):
        dask_trj, md_trj = self.slice_trajs(opt)

        c = list(itt.combinations(self.indices[:10], 3))
        function_test(dask_trj=dask_trj,
                      md_trj=md_trj,
                      funcname='compute_angles',
                      angle_indices=c,
                      opt=opt,
                      periodic=periodic)

    def test_compute_angles_wrong_indices(self):
        c = [[-1, 3000, 8]]
        with pytest.raises(ValueError) as err:
            dmd.compute_angles(traj=self.dask_traj,
                               angle_indices=c)
        assert 'angle_indices' in str(err.value)

    def test_compute_angles_empty_pairs(self):
        c = np.array([[1234, -9999, 8]])
        c = c[0:0]
        function_test(dask_trj=self.dask_traj,
                      md_trj=self.md_traj,
                      funcname='compute_angles',
                      angle_indices=c)



import pytest
from dask_traj.utils import (ensure_type,
                             lengths_and_angles_to_box_vectors,
                             box_vectors_to_lengths_and_angles)
import numpy as np
import itertools as itt
import mdtraj


class TestEnsureType(object):
    def setup(self):
        self.kwargs = {'dtype': np.int32,
                       'ndim': 1,
                       'name': 'test',
                       'add_newaxis_on_deficient_ndim': True}

    def teardown(self):
        pass

    def test_generators(self):
        test = itt.repeat(1, 3)  # This is a generator in python3
        b = ensure_type(test, **self.kwargs)
        assert all(b == np.array([1, 1, 1], dtype=np.int32))

    def test_scalar_conversion(self):
        test = 1
        self.kwargs['warn_on_cast'] = False
        b = ensure_type(test, **self.kwargs)
        assert all(b == np.array([1], dtype=np.int32))

    def test_type_error(self):
        test = 1.0
        self.kwargs.pop('add_newaxis_on_deficient_ndim')
        with pytest.raises(TypeError) as err:
            _ = ensure_type(test, **self.kwargs)
        assert 'must be numpy array' in str(err.value)

    def test_adding_ndim_to_arrays(self):
        test = np.array([1.0], dtype=np.int32)
        self.kwargs['ndim'] = 2
        b = ensure_type(test, **self.kwargs)
        assert all(b == np.array([[1.0]], dtype=np.int32))

    def test_raising_on_wrong_ndim_arrays(self):
        test = np.array([1.0], dtype=np.int32)
        self.kwargs['ndim'] = 2
        self.kwargs.pop('add_newaxis_on_deficient_ndim')
        with pytest.raises(ValueError) as err:
            _ = ensure_type(test, **self.kwargs)
        assert 'must be ndim' in str(err.value)

    def test_raising_on_wrong_length(self):
        test = np.array([1.0], dtype=np.int32)
        with pytest.raises(ValueError) as err:
            _ = ensure_type(test, length=3, **self.kwargs)
        assert 'must be length' in str(err.value)

    def test_wrong_shape_length(self):
        test = np.array([1.0], dtype=np.int32)
        with pytest.raises(ValueError) as err:
            _ = ensure_type(test, shape=(1, 3), **self.kwargs)
        assert 'must be shape' in str(err.value)

    def test_wrong_shape(self):
        test = np.array([1.0], dtype=np.int32)
        with pytest.raises(ValueError) as err:
            _ = ensure_type(test, shape=(3,), **self.kwargs)
        assert 'must be shape' in str(err.value)


class TestLengthAndAnglesToBoxVectors(object):
    def setup(self):
        self.kwargs = {'a_length': 1,
                       'b_length': 1,
                       'c_length': 1,
                       'alpha': 90,
                       'beta': 90,
                       'gamma': 90}

    def teardown(self):
        pass

    def test_warn_on_radian(self):
        self.kwargs['alpha'] = 1
        self.kwargs['beta'] = 2
        self.kwargs['gamma'] = 3

        with pytest.warns(Warning) as record:
            _ = lengths_and_angles_to_box_vectors(**self.kwargs)
        assert 'radian' in record[0].message.args[0]

    def test_messed_up_shape(self):
        self.kwargs['a_length'] = np.asarray([[1]])
        with pytest.raises(TypeError) as err:
            _ = lengths_and_angles_to_box_vectors(**self.kwargs)
        assert 'Shape is messed up' in str(err)

    def test_same_as_mdtraj(self):
        a = mdtraj.utils.lengths_and_angles_to_box_vectors(**self.kwargs)
        b = lengths_and_angles_to_box_vectors(**self.kwargs)
        assert (a[0] == b[0].compute()).all()
        assert (a[1] == b[1].compute()).all()
        assert (a[2] == b[2].compute()).all()


class TestBoxVectorsToLengthAndAngles(object):
    def setup(self):
        self.kwargs = {'a': np.array([1, 0, 0]),
                       'b': np.array([0, 1, 0]),
                       'c': np.array([0, 0, 1])}

    def teardown(self):
        pass

    def test_non_uniform_shape(self):
        self.kwargs['a'] = np.array([1, 0, 0, 0])
        with pytest.raises(TypeError) as err:
            _ = box_vectors_to_lengths_and_angles(**self.kwargs)
        assert "Shape is messed up" in str(err)

    def test_wrong_shape(self):
        self.kwargs['a'] = np.array([1, 0])
        self.kwargs['b'] = np.array([1, 0])
        self.kwargs['c'] = np.array([1, 0])

        with pytest.raises(TypeError) as err:
            _ = box_vectors_to_lengths_and_angles(**self.kwargs)
        assert "length 3" in str(err)

    def test_wrong_ndim(self):
        self.kwargs['a'] = np.array([[[1, 0, 0]]])
        self.kwargs['b'] = np.array([[[1, 0, 0]]])
        self.kwargs['c'] = np.array([[[1, 0, 0]]])

        with pytest.raises(ValueError) as err:
            _ = box_vectors_to_lengths_and_angles(**self.kwargs)
        assert "1d or 2d" in str(err)



import dask
import dask.array as da
from dask_traj.utils import ensure_type, wrap_da
from mdtraj.geometry import _geometry
from mdtraj.geometry.distance import (_distance, _distance_mic,
                                      _displacement, _displacement_mic)
import numpy as np


@dask.delayed
def _compute_distances_chunk(xyz, pairs, box=None, periodic=True, opt=True,
                             orthogonal=False):
    """Compute distances for a single chunk

    Parameters
    ----------
    xyz : ndarray of shape (any, any, 3)
        The xyz coordinates of the chunk
    pairs : array of shape (any, 2)
        The indices for which to compute an angle
    box : ndarray of shape (any, 3, 3)
        The box vectors of the chunk
    periodic : bool
        Wether to use the periodc boundary during the calculation.
    opt : bool, default=True
        Use an optimized native library to calculate distances. MDTraj's
        optimized SSE angle calculation implementation is 10-20x faster than
        the (itself optimized) numpy implementation.
    orthogonal : bool or da.bool
        Wether all angles are close to 90 degrees
    """

    xyz = ensure_type(xyz, dtype=np.float32, ndim=3, name='xyz',
                      shape=(None, None, 3), warn_on_cast=False,
                      cast_da_to_np=True)
    # Cast orthogonal to bool (incase we get a delayed bool)
    orthogonal = bool(orthogonal)
    if periodic and box is not None:
        if opt:
            out = np.empty((xyz.shape[0], pairs.shape[0]), dtype=np.float32)
            _geometry._dist_mic(xyz, pairs, box.transpose(0, 2, 1).copy(), out,
                                orthogonal)
            return out
        else:
            return _distance_mic(xyz, pairs, box.transpose(0, 2, 1),
                                 orthogonal)

    # Either there are no unitcell vectors or they dont want to use them
    if opt:
        out = np.empty((xyz.shape[0], pairs.shape[0]), dtype=np.float32)
        _geometry._dist(xyz, pairs, out)
        return out
    else:
        return _distance(xyz, pairs)


def compute_distances(traj, atom_pairs, periodic=True, **kwargs):
    """Daskified version of mdtraj.compute_distances().

    This mimics py:method:`mdtraj.compute_distances()` but returns the answer
    as a py:class:`dask.array` object

    Parameters
    ----------
    traj : :py:class:`dask_traj.Trajectory`
        The trajectory to compute the angles for.
    atom_pairs : array of shape(any, 2)
        The indices for which to compute an distances.
    periodic : bool
        Wether to use the periodc boundary during the calculation.
    opt : bool, default=True
        Use an optimized native library to calculate distances. MDTraj's
        optimized SSE angle calculation implementation is 10-20x faster than
        the (itself optimized) numpy implementation.

    Returns
    -------
    distances : dask.array, shape(n_frames, atom_pairs)
        Dask array with the delayed calculated distance for each item in
        atom_pairs for each frame.
    """
    xyz = traj.xyz
    length = len(xyz)
    atoms = len(atom_pairs)
    pairs = ensure_type(atom_pairs, dtype=np.int32, ndim=2, name='atom_pairs',
                        shape=(None, 2), warn_on_cast=False)
    if not np.all(np.logical_and(pairs < traj.n_atoms, pairs >= 0)):
        raise ValueError('atom_pairs must be between 0 and %d' % traj.n_atoms)
    if len(pairs) == 0:  # If pairs is an empty slice of an array
        return da.zeros((length, 0), dtype=np.float32)
    if periodic and traj._have_unitcell:
        box = ensure_type(traj.unitcell_vectors, dtype=np.float32, ndim=3,
                          name='unitcell_vectors', shape=(len(xyz), 3, 3),
                          warn_on_cast=False)
    else:
        box = None
        orthogonal = False
    lazy_results = []
    current_frame = 0
    for frames in xyz.chunks[0]:
        chunk_size = (frames, atoms)
        next_frame = current_frame+frames
        if box is not None:
            current_box = box[current_frame:next_frame]
            orthogonal = da.allclose(
                traj.unitcell_angles[current_frame:next_frame], 90)
        else:
            current_box = None
        lazy_results.append(wrap_da(_compute_distances_chunk, chunk_size,
                                    xyz=xyz[current_frame:next_frame],
                                    pairs=pairs,
                                    box=current_box,
                                    orthogonal=orthogonal,
                                    **kwargs))
        current_frame = next_frame
    max_result = da.concatenate(lazy_results)
    result = max_result[:length]
    return result


@dask.delayed
def _compute_displacements_chunk(xyz, pairs, box=None, periodic=True, opt=True,
                                 orthogonal=False):
    """Compute displacements for a single chunk

    Parameters
    ----------
    xyz : ndarray of shape (any, any, 3)
        The xyz coordinates of the chunk
    pairs : array of shape (any, 2)
        The indices for which to compute an angle
    box : ndarray of shape (any, 3, 3)
        The box vectors of the chunk
    periodic : bool
        Wether to use the periodc boundary during the calculation.
    opt : bool, default=True
        Use an optimized native library to calculate distances. MDTraj's
        optimized SSE angle calculation implementation is 10-20x faster than
        the (itself optimized) numpy implementation.
    orthogonal : bool or da.bool
        Wether all angles are close to 90 degrees
    """

    # Cast orthogonal to a bool, just incase we got a delayed object
    orthogonal = bool(orthogonal)
    xyz = ensure_type(xyz, dtype=np.float32, ndim=3, name='xyz',
                      shape=(None, None, 3), warn_on_cast=False,
                      cast_da_to_np=True)
    if periodic and box is not None:
        if opt:
            out = np.empty((xyz.shape[0], pairs.shape[0], 3), dtype=np.float32)
            _geometry._dist_mic_displacement(xyz, pairs,
                                             box.transpose(0, 2, 1).copy(),
                                             out,
                                             orthogonal)
            return out
        else:
            return _displacement_mic(xyz, pairs, box.transpose(0, 2, 1),
                                     orthogonal)

    # Either there are no unitcell vectors or they dont want to use them
    if opt:
        out = np.empty((xyz.shape[0], pairs.shape[0], 3), dtype=np.float32)
        _geometry._dist_displacement(xyz, pairs, out)
        return out
    else:
        return _displacement(xyz, pairs)


def compute_displacements(traj, atom_pairs, periodic=True, **kwargs):
    """Daskified version of mdtraj.geometry.compute_displacements

    This mimics py:method:`mdtraj.compute_displacements()` but returns the
    answer as a py:class:`dask.array` object

    Parameters
    ----------
    traj : :py:class:`dask_traj.Trajectory`
        The trajectory to compute the angles for.
    atom_pairs : array of shape(any, 2)
        The indices for which to compute an distances.
    periodic : bool
        Wether to use the periodc boundary during the calculation.
    opt : bool, default=True
        Use an optimized native library to calculate distances. MDTraj's
        optimized SSE angle calculation implementation is 10-20x faster than
        the (itself optimized) numpy implementation.

    Returns
    -------
    displacements : dask.array, shape(n_frames, atom_pairs, 3)
        Dask array with the delayed calculated displacements for each item in
        atom_pairs for each frame.
    """

    xyz = traj.xyz
    length = len(xyz)
    atoms = len(atom_pairs)
    pairs = ensure_type(atom_pairs, dtype=np.int32, ndim=2, name='atom_pairs',
                        shape=(None, 2), warn_on_cast=False)
    if not np.all(np.logical_and(pairs < traj.n_atoms, pairs >= 0)):
        raise ValueError('atom_pairs must be between 0 and %d' % traj.n_atoms)
    if len(pairs) == 0:  # If pairs is an empty slice of an array
        return da.zeros((length, 0, 3), dtype=np.float32)
    if periodic and traj._have_unitcell:
        box = ensure_type(traj.unitcell_vectors, dtype=np.float32, ndim=3,
                          name='unitcell_vectors', shape=(len(xyz), 3, 3),
                          warn_on_cast=False)
    else:
        box = None
        orthogonal = False
    lazy_results = []
    current_frame = 0
    for frames in xyz.chunks[0]:
        chunk_size = (frames, atoms, 3)
        next_frame = current_frame+frames
        if box is not None:
            current_box = box[current_frame:next_frame]
            orthogonal = da.allclose(
                traj.unitcell_angles[current_frame:next_frame], 90)
        else:
            current_box = None
        lazy_results.append(wrap_da(_compute_displacements_chunk, chunk_size,
                                    xyz=xyz[current_frame:next_frame],
                                    pairs=pairs,
                                    box=current_box,
                                    orthogonal=orthogonal,
                                    **kwargs))
        current_frame = next_frame
    max_result = da.concatenate(lazy_results)
    result = max_result[:length]
    return result


@dask.delayed
def _compute_center_of_mass_chunk(xyz, masses):
    """Compute center of mass coordinates for a single chunk

    Parameters
    ----------
    xyz : ndarray of shape (any, any, 3)
        The xyz coordinates of the chunk
    masses : ndarray of shape (n_atoms, )
        The masses of each atom
    """
    com = np.zeros((xyz.shape[0], 3))
    for i, x in enumerate(xyz):
        com[i, :] = x.astype('float64').T.dot(masses)
    return com


def compute_center_of_mass(traj):
    """Daskified version of mdtraj.geometry.compute_center_of_mass

    This mimics py:method:`mdtraj.compute_center_of_mass()` but returns the
    answer as a py:class:`dask.array` object

    Parameters
    ----------
    traj : :py:class:`dask_traj.Trajectory`
        The trajectory to compute the angles for.

    Returns
    -------
    com : dask.array, shape(n_frames, 3)
        Dask array with the delayed calculated Coordinates of center of mass
        for each frame.
    """

    xyz = traj.xyz
    length = len(xyz)
    masses = np.array([a.element.mass for a in traj.top.atoms])
    masses /= masses.sum()
    lazy_results = []
    current_frame = 0
    for frames in xyz.chunks[0]:
        chunk_size = (frames, 3)
        next_frame = current_frame+frames
        lazy_results.append(wrap_da(f=_compute_center_of_mass_chunk,
                                    chunk_size=chunk_size,
                                    xyz=xyz[current_frame:next_frame],
                                    masses=masses
                                    ))
        current_frame = next_frame
    max_result = da.concatenate(lazy_results)
    results = max_result[:length]
    return results


@dask.delayed
def _compute_center_of_geometry_chunk(xyz):
    """Compute center of geometry coordinates for a single chunk

    Parameters
    ----------
    xyz : ndarray of shape (any, any, 3)
        The xyz coordinates of the chunk
    """
    centers = np.zeros((xyz.shape[0], 3))
    for i, x in enumerate(xyz):
        centers[i, :] = x.astype('float64').T.mean(axis=1)
    return centers


def compute_center_of_geometry(traj):
    """Daskified version of mdtraj.geometry.compute_center_of_geometry

    This mimics py:method:`mdtraj.compute_center_of_geometry()` but returns the
    answer as a py:class:`dask.array` object

    Parameters
    ----------
    traj : :py:class:`dask_traj.Trajectory`
        The trajectory to compute the angles for.

    Returns
    -------
    com : dask.array, shape(n_frames, 3)
        Dask array with the delayed calculated Coordinates of center of
        geometry for each frame.
    """

    xyz = traj.xyz
    length = len(xyz)
    lazy_results = []
    current_frame = 0
    for frames in xyz.chunks[0]:
        chunk_size = (frames, 3)
        next_frame = current_frame+frames
        lazy_results.append(wrap_da(f=_compute_center_of_geometry_chunk,
                                    chunk_size=chunk_size,
                                    xyz=xyz[current_frame:next_frame],
                                    ))
        current_frame = next_frame
    max_result = da.concatenate(lazy_results)
    results = max_result[:length]
    return results


@dask.delayed
def _find_closest_contact(traj, group1, group2, frame=0, periodic=True):

    xyz = ensure_type(traj.xyz[frame], dtype=np.float32, ndim=2,
                      name='xyz', shape=(None, 3), warn_on_cast=False,
                      cast_da_to_np=True)
    atoms1 = ensure_type(group1, dtype=np.int32, ndim=1,
                         name='group1', warn_on_cast=False)
    atoms2 = ensure_type(group2, dtype=np.int32, ndim=1,
                         name='group2', warn_on_cast=False)
    if periodic and traj._have_unitcell:
        box = ensure_type(traj.unitcell_vectors, dtype=np.float32,
                          ndim=3, name='unitcell_vectors',
                          shape=(len(traj.xyz), 3, 3),
                          warn_on_cast=False)[frame]
    else:
        box = None
    ans = _geometry._find_closest_contact(xyz, atoms1, atoms2, box)
    return np.asarray(ans)


def find_closest_contact(traj, group1, group2, frame=0, periodic=True):
    """Daskified version fo mdtraj.find_closest_contact()

    Mimics mdtraj.find_closest_contact, but instead returns a delayed tuple

    Parameters
    ----------
    traj : Trajectory
        An dask-traj trajectory.
    group1 : ndarray, shape=(num_atoms), dtype=int
        The indices of atoms in the first group.
    group2 : ndarray, shape=(num_atoms), dtype=int
        The indices of atoms in the second group.
    frame : int, default=0
        The frame of the Trajectory to take positions from
    periodic : bool, default=True
        If `periodic` is True and the trajectory contains unitcell
        information, we will compute distances under the minimum image
        convention.
    Returns
    -------
    result : delayed tuple (int, int, float)
         A delayed tuple with the indices of the two atoms forming the closest
         contact, and the distance between them.
    """

    # This shouldn't be used, as it is the worst possible example of the
    # scaling (1 frame only).
    # Just here for completenes
    ans = wrap_da(f=_find_closest_contact,
                  chunk_size=(3,),
                  traj=traj,
                  group1=group1,
                  group2=group2,
                  frame=frame,
                  periodic=periodic)
    return dask.delayed(tuple)([int(ans[0]), int(ans[1]), float(ans[2])])



# Copy of various mdtraj utils that need to loosen, to work with dask
from __future__ import print_function, division
import warnings
import numpy as np
import collections
from mdtraj.utils.six.moves import zip_longest
import dask.array as da


class TypeCastPerformanceWarning(RuntimeWarning):
    pass


def ensure_type(val, dtype, ndim, name, length=None, can_be_none=False,
                shape=None, warn_on_cast=True,
                add_newaxis_on_deficient_ndim=False,
                cast_da_to_np=False):
    """Typecheck the size, shape and dtype of a numpy array, with optional
    casting.
    Parameters
    ----------
    val : {np.ndaraay, None}
        The array to check
    dtype : {nd.dtype, str}
        The dtype you'd like the array to have
    ndim : int
        The number of dimensions you'd like the array to have
    name : str
        name of the array. This is used when throwing exceptions, so that
        we can describe to the user which array is messed up.
    length : int, optional
        How long should the array be?
    can_be_none : bool
        Is ``val == None`` acceptable?
    shape : tuple, optional
        What should be shape of the array be? If the provided tuple has
        Nones in it, those will be semantically interpreted as matching
        any length in that dimension. So, for example, using the shape
        spec ``(None, None, 3)`` will ensure that the last dimension is of
        length three without constraining the first two dimensions
    warn_on_cast : bool, default=True
        Raise a warning when the dtypes don't match and a cast is done.
    add_newaxis_on_deficient_ndim : bool, default=False
        Add a new axis to the beginining of the array if the number of
        dimensions is deficient by one compared to your specification. For
        instance, if you're trying to get out an array of ``ndim == 3``,
        but the user provides an array of ``shape == (10, 10)``, a new axis
        will be created with length 1 in front, so that the return value is of
        shape ``(1, 10, 10)``.
    cast_da_to_np : bool, default=False
        Cast dask.arrays to np.arrays, this will trigger a computation on the
        dask array, which is needed to make them C-contigious in memory for
        several low-level mdtraj computations
    Notes
    -----
    The returned value will always be C-contiguous if it is an np.array or if
    cast_da_to_np is True.

    Returns
    -------
    typechecked_val : ndarray, None
        If `val=None` and `can_be_none=True`, then this will return None.
        Otherwise, it will return val (or a copy of val). If the dtype wasn't
        right, it'll be casted to the right shape. If the array was not
        C-contiguous, it'll be copied as well.
    """
    if can_be_none and val is None:
        return None

    if not isinstance(val, (np.ndarray, da.core.Array)):
        if isinstance(val, collections.Iterable):
            # If they give us an iterator, let's try...
            if isinstance(val, collections.Sequence):
                # sequences are easy. these are like lists and stuff
                val = np.array(val, dtype=dtype)
            else:
                # this is a generator...
                val = np.array(list(val), dtype=dtype)
        elif np.isscalar(val) and add_newaxis_on_deficient_ndim and ndim == 1:
            # special case: if the user is looking for a 1d array, and
            # they request newaxis upconversion, and provided a scalar
            # then we should reshape the scalar to be a 1d length-1 array
            val = np.array([val])
        else:
            raise TypeError(("%s must be numpy array. "
                             " You supplied type %s" % (name, type(val))))

    if warn_on_cast and val.dtype != dtype:
        warnings.warn("Casting %s dtype=%s to %s " % (name, val.dtype, dtype),
                      TypeCastPerformanceWarning)

    if not val.ndim == ndim:
        if add_newaxis_on_deficient_ndim and val.ndim + 1 == ndim:
            val = val[np.newaxis, ...]
        else:
            raise ValueError(("%s must be ndim %s. "
                              "You supplied %s" % (name, ndim, val.ndim)))

    if (isinstance(val, np.ndarray) or (
            cast_da_to_np and isinstance(val, da.core.Array))):
        val = np.ascontiguousarray(val, dtype=dtype)

    if length is not None and len(val) != length:
        raise ValueError(("%s must be length %s. "
                          "You supplied %s" % (name, length, len(val))))

    if shape is not None:
        # the shape specified given by the user can look like (None, None 3)
        # which indicates that ANY length is accepted in dimension 0 or
        # dimension 1
        sentenel = object()
        error = ValueError(("%s must be shape %s. You supplied  "
                            "%s" % (name, str(shape).replace('None', 'Any'),
                                    val.shape)))
        for a, b in zip_longest(val.shape, shape, fillvalue=sentenel):
            if a is sentenel or b is sentenel:
                # if the sentenel was reached, it means that the ndim didn't
                # match or something. this really shouldn't happen
                raise error
            if b is None:
                # the user's shape spec has a None in it, it matches anything
                continue
            if a != b:
                # check for equality
                raise error
    return val


def lengths_and_angles_to_box_vectors(a_length, b_length, c_length,
                                      alpha, beta, gamma):
    """Convert from the lengths/angles of the unit cell to the box
    vectors (Bravais vectors). The angles should be in degrees.

    Mimics mdtraj.core.unitcell.lengths_and_angles_to_box_vectors()

    Parameters
    ----------
    a_length : scalar or ndarray
        length of Bravais unit vector **a**
    b_length : scalar or ndarray
        length of Bravais unit vector **b**
    c_length : scalar or ndarray
        length of Bravais unit vector **c**
    alpha : scalar or ndarray
        angle between vectors **b** and **c**, in degrees.
    beta : scalar or ndarray
        angle between vectors **c** and **a**, in degrees.
    gamma : scalar or ndarray
        angle between vectors **a** and **b**, in degrees.

    Returns
    -------
    a : dask.array
        If the inputs are scalar, the vectors will one dimensional (length 3).
        If the inputs are one dimension, shape=(n_frames, ), then the output
        will be (n_frames, 3)
    b : dask.array
        If the inputs are scalar, the vectors will one dimensional (length 3).
        If the inputs are one dimension, shape=(n_frames, ), then the output
        will be (n_frames, 3)
    c : dask.array
        If the inputs are scalar, the vectors will one dimensional (length 3).
        If the inputs are one dimension, shape=(n_frames, ), then the output
        will be (n_frames, 3)

    This code is adapted from gyroid, which is licensed under the BSD
    http://pythonhosted.org/gyroid/_modules/gyroid/unitcell.html
    """
    # Fix for da that requires angles and lengths to be arrays
    lengths = [a_length, b_length, c_length]
    for i, e in enumerate(lengths):
        # Use python logic shortcutting to not compute dask Arrays
        if not isinstance(e, da.core.Array) and np.isscalar(e):
            lengths[i] = np.array([e])
    a_length, b_length, c_length = tuple(lengths)

    angles = [alpha, beta, gamma]
    for i, e in enumerate(angles):
        if not isinstance(e, da.core.Array) and np.isscalar(e):
            angles[i] = np.array([e])
    alpha, beta, gamma = tuple(angles)

    if da.all(alpha < 2*np.pi) and (da.all(beta < 2*np.pi)
                                    and da.all(gamma < 2*np.pi)):
        warnings.warn('All your angles were less than 2*pi.'
                      ' Did you accidentally give me radians?')

    alpha = alpha * np.pi / 180
    beta = beta * np.pi / 180
    gamma = gamma * np.pi / 180

    a = da.stack([a_length, da.zeros_like(a_length), da.zeros_like(a_length)])
    b = da.stack([b_length*da.cos(gamma),
                  b_length*da.sin(gamma),
                  da.zeros_like(b_length)])
    cx = c_length*da.cos(beta)
    cy = c_length*(da.cos(alpha) - da.cos(beta)*da.cos(gamma)) / da.sin(gamma)
    cz = da.sqrt(c_length*c_length - cx*cx - cy*cy)
    c = da.stack([cx, cy, cz])
    if not a.shape == b.shape == c.shape:
        raise TypeError('Shape is messed up.')

    # Make sure that all vector components that are _almost_ 0 are set exactly
    # to 0
    tol = 1e-6
    a[da.logical_and(a > -tol, a < tol)] = 0.0
    b[da.logical_and(b > -tol, b < tol)] = 0.0
    c[da.logical_and(c > -tol, c < tol)] = 0.0

    return a.T, b.T, c.T


def box_vectors_to_lengths_and_angles(a, b, c):
    """Convert box vectors into the lengths and angles defining the box.

    Addapted from mdtraj.utils.unitcell.box_vectors_to_lengths_and_angles()
    Parameters
    ----------
    a : np.ndarray
        the vector defining the first edge of the periodic box (length 3), or
        an array of this vector in multiple frames, where a[i,:] gives the
        length 3 array of vector a in each frame of a simulation
    b : np.ndarray
        the vector defining the second edge of the periodic box (length 3), or
        an array of this vector in multiple frames, where b[i,:] gives the
        length 3 array of vector a in each frame of a simulation
    c : np.ndarray
        the vector defining the third edge of the periodic box (length 3), or
        an array of this vector in multiple frames, where c[i,:] gives the
        length 3 array of vector a in each frame of a simulation

    Returns
    -------
    a_length : dask.array
        length of Bravais unit vector **a**
    b_length : dask.array
        length of Bravais unit vector **b**
    c_length : dask.array
        length of Bravais unit vector **c**
    alpha : dask.array
        angle between vectors **b** and **c**, in degrees.
    beta : dask.array
        angle between vectors **c** and **a**, in degrees.
    gamma : dask.array
        angle between vectors **a** and **b**, in degrees.
    """
    if not a.shape == b.shape == c.shape:
        raise TypeError('Shape is messed up.')
    if not a.shape[-1] == 3:
        raise TypeError('The last dimension must be length 3')
    if not (a.ndim in [1, 2]):
        raise ValueError('vectors must be 1d or 2d (for a vectorized '
                         'operation on multiple frames)')
    last_dim = a.ndim-1

    a_length = (da.sum(a*a, axis=last_dim))**(1/2)
    b_length = (da.sum(b*b, axis=last_dim))**(1/2)
    c_length = (da.sum(c*c, axis=last_dim))**(1/2)

    # we allow 2d input, where the first dimension is the frame index
    # so we want to do the dot product only over the last dimension
    alpha = da.arccos(da.einsum('...i, ...i', b, c) / (b_length * c_length))
    beta = da.arccos(da.einsum('...i, ...i', c, a) / (c_length * a_length))
    gamma = da.arccos(da.einsum('...i, ...i', a, b) / (a_length * b_length))

    # convert to degrees
    alpha = alpha * 180.0 / np.pi
    beta = beta * 180.0 / np.pi
    gamma = gamma * 180.0 / np.pi

    return a_length, b_length, c_length, alpha, beta, gamma


def wrap_da(f, chunk_size, **kwargs):
    """Convenience function to wrap dask.array.from_delayed()"""
    return da.from_delayed(f(**kwargs), dtype=np.float32, shape=chunk_size)
